// -*- c++ -*-
//
// SUMMARY:      
// USAGE:        
//
// AUTHOR:       Christophe Prud'homme
// ORG:          
// E-MAIL:       prudhomm@ann.jussieu.fr
//
// ORIG-DATE:     6-Jan-96 at 15:08:34
// LAST-MOD:      9-Feb-96 at 23:45:23 by Prud'homme Christophe
//
// DESCRIPTION:  
// DESCRIP-END.
//
// $Id: comm.cxx,v 1.1.1.1 1996/11/02 02:48:19 prudhomm Exp $
//
// $Log: comm.cxx,v $
// Revision 1.1.1.1  1996/11/02 02:48:19  prudhomm
// Importing files from release 3.3
//
//

#if __GNUC__
#pragma implementation
#endif /* __GNUC__ */

#include <comm.hxx>

char SOCKNAME[256];

comm::comm (char *_id,int _type, int _port):type(_type), port(_port),
  id(new char[strlen(_id)+1])
{
  int isize;
  char buf[256];
  strcpy (id, _id);
  sprintf (buf, "/tmp/gfemlog-%s", id);
  ferr.open(buf);
  
  // socket creation
  if ((sid  = socket(AF_UNIX, type, 0)) == -1)
    {
      ferr << id << "  Unable to create the socket of type: " << type << endl;
      perror("socket");
      exit(-2);
    }
  sprintf(SOCKNAME, "/tmp/comm-gfem");

  // preparation data for attachment
  address.sun_family = AF_UNIX;
  strncpy (address.sun_path, SOCKNAME,
           sizeof(struct sockaddr_un)-sizeof(short));

  // ask for attachment
  if (bind (sid, (struct sockaddr *)&address, sizeof (struct sockaddr_un)) == -1)
      {
#if defined(_HANDLE_EXCEPTIONS)
        throw ("  Unable to bind the socket");
#else
        perror ("bind");
        cclose (port);
        init (_id, _type, _port);
#endif
      }
  cid = sid;
  isize = BUFSIZE;
  if (setsockopt (sid, SOL_SOCKET, SO_SNDBUF, &isize, sizeof(double)) == -1)
    perror ("setsockopt failed");
  if (setsockopt (sid, SOL_SOCKET, SO_RCVBUF, &isize, sizeof(double)) == -1)
    perror ("setsockopt failed");
}
void
comm::init (char *_id,int _type, int _port)
{
  int isize;
  char error[256];
  char buf[256];
  id = new char[strlen(_id)+1];
  strcpy (id, _id);
  sprintf (buf, "/tmp/gfemlog-%s", id);
  ferr.open(buf);
  type = _type;
  port = _port;
  
  sprintf(SOCKNAME, "/tmp/comm-gfem");
  
  // socket creation
  ferr << "Creating the socket...";
  if ((sid  = socket(AF_UNIX, type, 0)) == -1)
    {
      ferr << id << "  Unable to create the socket of type: " << type << endl;
      perror("socket");
      exit(-2);
    }
  ferr << "Done.\n";
  
  // preparation data for attachment
  address.sun_family = AF_UNIX;
  strncpy (address.sun_path, SOCKNAME,
           sizeof(struct sockaddr_un)-sizeof(short));
  
  // ask for attachment
  if (port == 0) // Server
    {
      ferr << "Binding the socket...";
      if (bind (sid, (struct sockaddr *)&address, sizeof (struct sockaddr_un)) == -1)
	{
	  sprintf (error, "<Master> bind::");
	  perror (error);
	  cclose (port);
	}
      ferr << "Done.\n";
    }
  cid = sid;
  isize = BUFSIZE;
  if (setsockopt (sid, SOL_SOCKET, SO_SNDBUF, &isize, sizeof(double)) == -1)
    perror ("setsockopt failed");
  if (setsockopt (sid, SOL_SOCKET, SO_RCVBUF, &isize, sizeof(double)) == -1)
    perror ("setsockopt failed");
}
//
// Communications
//
int
comm::caccept ()
{
  int sz = sizeof(address);

  ferr << "Accepting the connection...";
  while (1)
    {
      cid = accept(sid, (struct sockaddr *)&address, &sz);
      if (cid == -1 && errno == EINTR) // signal received , probably SIGCHLD
        continue;
      if (cid == -1) // error
        {
          perror ("accept:: ");
          cclose(0);
          exit (-2);
        }
      break;
    }
  ferr << "Done.\n";
  return 1;
}
void 
comm::cconnect ()
{
  // connect on the server
  ferr << "Connecting to the server...";
  sleep(2);
  if (connect(sid, (struct sockaddr *)&address, sizeof(struct sockaddr_un)) == -1)
    {
      perror ("connect:: ");
      cclose(1);
      exit (-2);
    }
  ferr << "Done.\n";
}
//
// receive
//
int 
comm::crecv (int *ptrint, int sz)
{
  return read (cid, ptrint, sz*sizeof(int));
}

int
comm::crecv (char *&tab)
{
  int sz,ret;
  crecv (sz);
  if (sz)
    {
      tab = new char[sz+1];
      ret = read (cid, tab, sz);
      tab[sz] = 0;
      return ret;
    }
  return sz;
}
int 
comm::crecv (float *ptrfloat, int sz)
{
  return read (cid, ptrfloat, sz*sizeof(float));
}

int 
comm::crecv (double *ptrdouble, int sz)
{
  return read (cid, ptrdouble, sz*sizeof(double));
}
int 
comm::crecv (rpoint *ptr, int sz)
{
  int cc;
  //sz *= sizeof(rpoint);
  while (sz > 0)
    {
      cc = read (cid, ptr, sz);
      if (cc == -1)
        {
          perror ("comm:crecv<rpoint>");
          cclose (port);
          exit(-1);
        }

      ptr += cc;
      sz -= cc;
    }
  return sz;
  //return read (cid, ptr, sz*sizeof(rpoint));
}
int 
comm::crecv (triangle *ptr, int sz)
{
  int cc;
  //  sz *= sizeof(triangle);
  while (sz > 0)
    {
      cc = read (cid, ptr, sz);
      if (cc == -1)
        {
          perror ("comm:crecv<triangle>");
          cclose (port);
          exit(-1);
        }
      ptr += cc;
      sz -= cc;
    }
  return sz;
  //return read (cid, ptr, sz*sizeof(triangle));
}
int 
comm::crecv (parameter& ptr)
{
  crecv (ptr.tinterp);
  crecv (ptr.err0);
  crecv (ptr.aniso);
  crecv (ptr.hwall);
  crecv (ptr.refwall);
  crecv (ptr.angulo);
  crecv (ptr.nit);
  crecv (ptr.segur);
  crecv (ptr.reg_ini);
  crecv (ptr.reg_fin);
  crecv (ptr.rel_mtr);
  crecv (ptr.fluid_NS);
  return 0;
}
//
// return send
//
int 
comm::csend (int *ptrint, int sz)
{
  return write (cid, ptrint, sz*sizeof(int));
}
int
comm::csend  (char *tab) 
{
  int sz = strlen (tab);
 
  csend (sz);
  if (sz)
    {
      return write (cid, tab, sz);
    }
  return sz;
}
int 
comm::csend (float *ptrfloat, int sz)
{
  return write (cid, ptrfloat, sz*sizeof(float));
}
int 
comm::csend (double *ptrdouble, int sz)
{
  return write (cid, ptrdouble, sz*sizeof(double));
}
int 
comm::csend (rpoint *ptr, int sz)
{
  int cc;
  //sz *= sizeof(rpoint);
  while (sz > 0)
    {
      cc = write (cid, ptr, sz);
      if (cc == -1)
        {
          perror ("comm:csend<rpoint>");
          cclose (port);
          exit(-1);
        }
      ptr += cc;
      sz -= cc;
    }
  return sz;
  //  return write (cid, ptr, sz*sizeof(rpoint));
}
int 
comm::csend (triangle *ptr, int sz)
{
  int cc;
  //sz *= sizeof(triangle);
  while (sz > 0)
    {
      cc = write (cid, ptr, sz);
      if (cc == -1)
        {
          perror ("comm:csend<triangle>");
          cclose (port);
          exit(-1);
        }
      ptr += cc;
      sz -= cc;
    }
  return sz;
  //return write (cid, ptrtr, sz*sizeof(triangle));
}
int 
comm::csend (parameter& ptr)
{
  csend (ptr.tinterp);
  csend (ptr.err0);
  csend (ptr.aniso);
  csend (ptr.hwall);
  csend (ptr.refwall);
  csend (ptr.angulo);
  csend (ptr.nit);
  csend (ptr.segur);
  csend (ptr.reg_ini);
  csend (ptr.reg_fin);
  csend (ptr.rel_mtr);
  csend (ptr.fluid_NS);
  return 0;
}
