// -*-   c++ -*-
//
// SUMMARY:      
// USAGE:        
//
// AUTHOR:       Christophe Prud'homme
// ORG:          Numrica Inc
// E-MAIL:        prudhomm@ann.jussieu.fr
//
// ORIG-DATE:     5-Jan-96 at 15:06:15
// LAST-MOD:     11-Sep-96 at 12:42:06 by Christophe Prudhomme
//
// DESCRIPTION:  
// DESCRIP-END.
//
// $Id: comm.hxx,v 1.1.1.1 1996/11/02 02:48:19 prudhomm Exp $
//
// $Log: comm.hxx,v $
// Revision 1.1.1.1  1996/11/02 02:48:19  prudhomm
// Importing files from release 3.3
//
//

#ifndef  COMM_H
#if __GNUC__
#pragma interface
#endif /* __GNUC__ */
#define COMM_H 1

#include <stdlib.h>
#include <unistd.h>
#include <iostream.h>
#include <fstream.h>
#include <string.h>

#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#if defined(HAVE_CONFIG_H)
# include <config.h>
#endif /* HAVE_CONFIG_H */

#if defined(HAVE_ERRNO_H)
# include <errno.h>
#endif /* HAVE_ERRNO_H */


#if !defined(TRIANGULATION_TYPEDEFS)
#define TRIANGULATION_TYPEDEFS
typedef struct
{
 float x;
 float y;
}rpoint;

typedef long triangle[3];

typedef struct
{
 long np, nt;
 rpoint *rp;
 triangle *tr;
 int* ngt;
 int* ng;
}triangulation;
#endif /* TRIANGULATION_TYPEDEFS */

#include <params.hxx>

#define INT 0
#define CHAR 1
#define FLOAT 2
#define DOUBLE 3
#define RPOINT 4
#define TRIANGLE 5

#define BUFSIZE 30000

extern char SOCKNAME[256];

class comm
{
private:
  int sid;    // socket id : server
  int cid;   //  socket id : client
  int type; // type of the socket
  int port; // port of the socket
  char *id; // string identifier of the socket
  int size[10]; // size of each 

  struct hostent *hp;
  struct sockaddr_un address; // attachment address

  ofstream ferr;

public:
  // constructors and destructors
  comm ();
  comm (char *, int = SOCK_STREAM, int = 0);
  ~comm();
  void cclose (int);
  void init (char *, int, int);

  // communications
  int clisten ();    // SERVER 1 = ok -1 = error
  int caccept();     // SERVER -2 = SIGCHLD -1 = error 1 = ok
  void cconnect ();// CLIENT

  // send and receive
  int crecv (int &);
  int crecv (int *, int);
  int crecv (char &);
  int crecv (char * &);
  int crecv (float&);
  int crecv (float *, int);
  int crecv (double&);
  int crecv (double*, int);
  int crecv (rpoint*, int);
  int crecv (rpoint &);
  int crecv (triangle *, int);
  int crecv (triangle &);
  int crecv (parameter&);
  
  int csend (int);
  int csend (int *, int);
  int csend (char);
  int csend (char *);
  int csend (float);
  int csend (float *, int);
  int csend (double);
  int csend (double *, int);
  int csend (rpoint *, int);
  int csend (rpoint);
  int csend (triangle *, int);
  int csend (triangle);
  int csend (parameter&);

  // debugging infos
  void print ();
};
inline
comm::comm ():sid(0)
{
}


inline 
void comm::cclose(int who)
{
  if (sid)
    switch (who)
      {
      case 0:
        ferr << "<Master> Close the socket\n";
        unlink(SOCKNAME);
        close (sid);
        sid = 0;
        break;
      case 1:
	//        unlink(SOCKNAME);
        ferr << "<Slave> Close the socket\n";
        close (sid);
        sid  = 0;
        break;
      }
}
inline 
comm::~comm()
{
  cclose (port);
}
// comunications
inline int
comm::clisten ()
{
  // open service
  ferr << "Listening to the socket...";
  if (listen(sid, 0) == -1)
    {
      perror("listen");
      cclose(0);
      return -1;
    }
  ferr <<"Done\n";
  return 1;
}

//
// receive
//
inline int 
comm::crecv (int &ptrint)
{
  return read (cid, &ptrint, sizeof(int));
}
inline int
comm::crecv (char &ptrchar)
{ 
  return read (cid, &ptrchar, sizeof(char));
}

inline int 
comm::crecv (float &ptrfloat)
{
  return read (cid, &ptrfloat, sizeof(float));
}

inline int 
comm::crecv (double &ptrdouble)
{
  return read (cid, &ptrdouble, sizeof(double));
}

inline int 
comm::crecv (rpoint & ptr)
{
  return read (cid, &ptr, sizeof(rpoint));
}
inline int 
comm::crecv (triangle & ptr)
{
  crecv ((int&)ptr[0]);  crecv ((int&)ptr[1]);  crecv ((int&)ptr[2]);
  return 1;
  //read (cid, &ptr, sizeof(triangle));
}

//
// send
//

inline int 
comm::csend (int ptrint)
{
  return write (cid, &ptrint, sizeof(int));
}

inline int
comm::csend  (char tab) 
{
  return write (cid, &tab, sizeof(char));
}

inline int 
comm::csend (float ptrfloat)
{
  return write (cid, &ptrfloat, sizeof(float));
}

inline int 
comm::csend (double ptrdouble)
{
  return write (cid, &ptrdouble, sizeof(double));
}
inline int 
comm::csend (rpoint ptr)
{
  return write (cid, &ptr, sizeof(rpoint));
}
inline int 
comm::csend (triangle ptr)
{
  int tot;
  tot = csend ((int)ptr[0]); 
  tot += csend ((int)ptr[1]);
  tot += csend ((int)ptr[2]);
  return tot;
  //return write (cid, &ptr, sizeof(triangle));
}

// debugging infos
inline void
comm::print ()
{
  ferr << id << "  sid = " << sid << endl;
}
#endif /* COMM_H */




